/************************************/
/*        A-アルゴリズム            */
/*             a.c                  */
/************************************/

#include <stdio.h>

/*   記号定数                      */
#define LIMITL 256 /*配列最大領域*/
#define TRUE 1 /*真（論理値）*/
#define START 1/*スタートの節点番号*/
#define GOAL 999 /*ゴールの節点番号*/

/*     節点を表現する構造体        */
struct node{
 int nodeid ;/*節点の番号*/
 int parentid ;/*接続元の節点番号*/
 double value ;/*評価の推定値*/
} ;
/* オープンリストとクローズドリスト*/
 struct node openlist[LIMITL] ;
 int openlistep=0 ;
 struct node closedlist[LIMITL] ;
 int closedlistep=0 ;
/*各接点の評価の推定値*/
double h[]={0,5.7,4.5,2.8,4.1,2,4,4,0} ;

/*   関数のプロトタイプの宣言      */
/*リストの初期化*/
void initlist() ;
/*リストの状態の出力*/
void printlist() ;
/*節点の展開*/
void expand(struct node cnode) ;
/*先頭要素をクローズドリストへ移動*/
void movetofirst() ;
/*オープンリストの先頭から削除*/
void removefirst() ; 
/*ＩＤの重複チェック*/
int ccheck(struct node cnode
      ,int id,int v) ; /*クローズドリスト*/
int ocheck(struct node cnode
      ,int id,int v) ; /*オープンリスト*/
/*ゴールへの経路を出力*/
void printroute(int id) ; 
/*オープンリストの昇順整列*/
void sortopenlist() ; 
/* 下請け関数　節点の比較関数*/
int cmp(const void *a,const void *b) ;


/**********************/
/*     main()関数     */
/**********************/
int main()
{
 struct node currentnode ;

 /*初期化*/
 initlist() ;
 printlist() ;
 /*探索の本体*/
 while(TRUE){
  if(openlistep==0){
   /*ゴールが見つからなかった*/
   printf("ゴールは見つかりませんでした\n") ;
   break ;
  }
  /*オープンリストの先頭がゴールなら終了*/
  if(openlist[0].nodeid==GOAL){
   printf("\nゴールを見つけました\n") ;
   printf("%d[%d]",openlist[0].nodeid,
                  openlist[0].parentid) ;
   printroute(openlist[0].parentid) ;
   break ;
  } 
  /*オープンリストの先頭要素を取り出す*/
  currentnode=openlist[0] ;
  removefirst() ; /*オープンリストの先頭から削除*/
  /*オープンリストの先頭要素を展開*/
  expand(currentnode) ;

  /*クローズドリストへ移動*/
  closedlist[closedlistep++]=currentnode  ;

  /*オープンリストの昇順整列*/
  sortopenlist() ;

  printlist() ;

 }
 
 return 0 ;
}
/**************************/
/*  sortopenlist()関数    */
/*オープンリストの昇順整列*/
/**************************/
void sortopenlist()
{
 qsort(openlist,openlistep,sizeof(struct node),cmp) ;
} 

/* 下請け関数　節点の比較関数*/
int cmp(const void *a,const void *b)
{
 struct node *x,*y ;
 x=(struct node *)a ;
 y=(struct node *)b ;
 if((x->value)<(y->value)) return -1 ;
 else if((x->value)>(y->value)) return 1 ;
 else return 0 ;
}

/**************************/
/*  printroute()関数      */
/*  ゴールへの経路を出力  */
/**************************/
void printroute(int id)  
{
 int i;
 /*節点番号idの接点を探す*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id){
   printf("<-%d[%d]",closedlist[i].nodeid,
                     closedlist[i].parentid) ;
   break ;
  }
 /*次の接点を出力*/
 if(closedlist[i].parentid!=0)
   /*スタートでなければ出力を続行*/ 
   printroute(closedlist[i].parentid) ;
 printf("\n") ;

}

/**************************/
/*    ccheck()関数        */
/*  ＩＤの重複チェック    */
/*　クローズドリスト      */
/**************************/
int ccheck(struct node cnode
                 ,int id,int v)  
{
 int i,j;

 /*クローズドリストと比較，良い方を残す*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id){
   if(closedlist[i].value<cnode.value+v){
    /*クローズドリストの方が良い*/
    return TRUE ;
   }
   else{
   /*クローズドリストから削除*/
    for(j=i;j<closedlistep;++j){
     closedlist[j]=closedlist[j+1] ;
    }
    --closedlistep ;
    }
  }
 return 0 ;
} 



/**************************/
/*    ocheck()関数        */
/*  ＩＤの重複チェック    */
/*　オープンリスト        */
/**************************/
int ocheck(struct node cnode
                 ,int id,int v)  
{
 int i,j;

 /*オープンリストと比較，良い方を残す*/
 for(i=0;i<openlistep;++i)
  if(openlist[i].nodeid==id){
   if(openlist[i].value<cnode.value+v){
    /*オープンリストの方が良い*/
    return TRUE ;
   }
   else{
   /*オープンリストから削除*/
    for(j=i;j<openlistep;++j){
     openlist[j]=openlist[j+1] ;
    }
    --openlistep ;
    }
  }
 return 0 ;
} 


/**************************/
/*   removefirst()関数    */
/*オープンリストの先頭削除*/
/**************************/
void removefirst()
{
 int i;
 for(i=0;i<openlistep;++i)
  openlist[i]=openlist[i+1] ;
 --openlistep ;
} 


/**********************/
/*  movefirst()関数   */
/*  先頭要素の移動    */
/**********************/
void movetofirst()
{
 closedlist[closedlistep++]=openlist[0]  ;
}

/**********************/
/*   expand()関数     */
/*   節点の展開       */
/**********************/
void expand(struct node cnode)
{
 /*接続関係の情報*/
 double tree[][10]={
  {1,2,2,4,3},
  {2,3,2,6,2},
  {3,2,2,5,2,999,2.8},
  {4,3,2.8},
  {5,7,2,999,2},
  {0} /*情報の終わり*/
 } ;
 int i=0 ;
 int j ;

 while(tree[i][0]!=0){
  if(tree[i][0]==cnode.nodeid){
   /*展開*/
   for(j=1;tree[i][j]!=0;j+=2){
    if(ocheck(cnode,tree[i][j], /*重複チェック　オープンリスト*/
       tree[i][j+1]+h[(int)tree[i][j]])!=TRUE
     && (ccheck(cnode,tree[i][j],/*重複チェック　クローズドリスト*/
       tree[i][j+1]+h[(int)tree[i][j]])!=TRUE)){
     openlist[openlistep].nodeid=tree[i][j] ;
     openlist[openlistep].parentid=cnode.nodeid ;
     openlist[openlistep++].value
      =cnode.value+tree[i][j+1]+h[(int)tree[i][j]] ;
    }
   }
   break ;
  }
  ++i ;
 }

}

/**********************/
/*  initlist()関数    */
/*  リストの初期化    */
/**********************/
void initlist()
{
 /*オープンリストの初期化*/
 openlist[0].nodeid=START ;
 openlist[0].parentid=0 ;
 ++openlistep ;
}

/**********************/
/*  printlist()関数   */
/* リストの状態の出力 */
/**********************/
void printlist()
{
 int i ;

 /*オープンリストの出力*/
 printf("\nオープンリスト　　") ;
 for(i=0;i<openlistep;++i)
  printf("%d[%d,%.1lf],",openlist[i].nodeid,
                openlist[i].parentid,
                openlist[i].value) ;
 printf("\n") ;

 /*クローズドリストの出力*/
 printf("クローズドリスト　") ;
 for(i=0;i<closedlistep;++i)
  printf("%d[%d],",closedlist[i].nodeid,
                closedlist[i].parentid) ;
 printf("\n") ;
} 

