/************************************/
/*         縦型探索プログラム       */
/*         depth.c                  */
/************************************/

#include <stdio.h>

/*   記号定数                      */
#define LIMITL 256 /*配列最大領域*/
#define TRUE 1 /*真（論理値）*/
#define START 1/*スタートの節点番号*/
#define GOAL 999 /*ゴールの節点番号*/

/*     節点を表現する構造体        */
struct node{
 int nodeid ;/*節点の番号*/
 int parentid ;/*接続元の節点番号*/
} ;
/* オープンリストとクローズドリスト*/
 struct node openlist[LIMITL] ;
 int openlistep=0 ;
 struct node closedlist[LIMITL] ;
 int closedlistep=0 ;

/*   関数のプロトタイプの宣言      */
/*リストの初期化*/
void initlist() ;
/*リストの状態の出力*/
void printlist() ;
/*節点の展開*/
void expand(int id) ;
/*オープンリストの先頭から削除*/
void removefirst() ; 
/*ＩＤの重複チェック*/
int check(int id) ; 
/*ゴールへの経路を出力*/
void printroute(int id) ; 
/*オープンリストからの重複削除*/
void checkopenlist(int id) ; 
/*オープンリストの先頭を空ける*/
void addopenlist() ; 


/**********************/
/*     main()関数     */
/**********************/
int main()
{
 struct node currentnode ;

 /*初期化*/
 initlist() ;
 printlist() ;
 /*探索の本体*/
 while(TRUE){
  if(openlistep==0){
   /*ゴールが見つからなかった*/
   printf("ゴールは見つかりませんでした\n") ;
   break ;
  }
  /*オープンリストの先頭がゴールなら終了*/
  if(openlist[0].nodeid==GOAL){
   printf("\nゴールを見つけました\n") ;
   printf("%d[%d]",openlist[0].nodeid,
                  openlist[0].parentid) ;
   printroute(openlist[0].parentid) ;
   break ;
  } 
  /*オープンリストの先頭要素を取り出す*/
  currentnode=openlist[0] ;
  removefirst() ; /*オープンリストの先頭から削除*/
  /*オープンリストの先頭要素を展開*/
  expand(currentnode.nodeid) ;

  /*クローズドリストへ移動*/
  closedlist[closedlistep++]=currentnode  ;

  printlist() ;

 }
 
 return 0 ;
}

/**************************/
/*  printroute()関数      */
/*  ゴールへの経路を出力  */
/**************************/
void printroute(int id)  
{
 int i;
 /*節点番号idの接点を探す*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id){
   printf("<-%d[%d]",closedlist[i].nodeid,
                     closedlist[i].parentid) ;
   break ;
  }
 /*次の接点を出力*/
 if(closedlist[i].parentid!=0)
   /*スタートでなければ出力を続行*/ 
   printroute(closedlist[i].parentid) ;
 printf("\n") ;

}

/**************************/
/*     check()関数        */
/*  ＩＤの重複チェック    */
/**************************/
int check(int id)  
{
 int i;
 int res=0 ;

 /*オープンリストとの重複チェック*/
 /*縦型探索では不要*/
/* for(i=0;i<openlistep;++i)
  if(openlist[i].nodeid==id)
   res=TRUE ;*/
 /*クローズドリストとの重複チェック*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id)
   res=TRUE ;
 return res ;
} 

/**************************/
/* checkopenlist()関数    */
/*オープンリストの重複削除*/
/**************************/
void checkopenlist(int id)  
{
 int i,j;

 for(i=1;i<openlistep;++i)
  if(openlist[i].nodeid==id){
   for(j=i;j<openlistep;++j)
    openlist[j]=openlist[j+1] ;
   --openlistep ;
   --i ;
  }
} 

/**************************/
/*   removefirst()関数    */
/*オープンリストの先頭削除*/
/**************************/
void removefirst()
{
 int i;
 for(i=0;i<openlistep;++i)
  openlist[i]=openlist[i+1] ;
 --openlistep ;
} 

/**********************/
/*   expand()関数     */
/*   節点の展開       */
/**********************/
void expand(int id)
{
 /*接続関係の情報*/
 int tree[][5]={
  {1,2,3,5},
  {2,4},
  {3,6,7},
  {5,4,8},
  {6,5,7},
  {8,7,999},
  {0} /*情報の終わり*/
 } ;
 int i=0 ;
 int j ;

 while(tree[i][0]!=0){
  if(tree[i][0]==id){
   /*展開*/
   for(j=1;tree[i][j]!=0;++j){
    if(check(tree[i][j])!=TRUE){/*重複チェック*/
     /*オープンリストの先頭を空ける*/
     addopenlist() ;
     /*オープンリストの先頭に追加*/
     openlist[0].nodeid=tree[i][j] ;
     openlist[0].parentid=id ;
    }
    /*オープンリストに重複があれば削除*/
    /*（縦型探索固有の処理）          */
    checkopenlist(tree[i][j]) ;
   }
   break ;
  }
  ++i ;
 }

}

/**************************/
/*   addopenlist()関数    */
/*オープンリストの先頭追加*/
/**************************/
void addopenlist()
{
 int i;
 for(i=openlistep;i>0;--i)
  openlist[i]=openlist[i-1] ;
 ++openlistep ;
} 

/**********************/
/*  initlist()関数    */
/*  リストの初期化    */
/**********************/
void initlist()
{
 /*オープンリストの初期化*/
 openlist[0].nodeid=START ;
 openlist[0].parentid=0 ;
 ++openlistep ;
}

/**********************/
/*  printlist()関数   */
/* リストの状態の出力 */
/**********************/
void printlist()
{
 int i ;

 /*オープンリストの出力*/
 printf("\nオープンリスト　　") ;
 for(i=0;i<openlistep;++i)
  printf("%d[%d],",openlist[i].nodeid,
                openlist[i].parentid) ;
 printf("\n") ;

 /*クローズドリストの出力*/
 printf("クローズドリスト　") ;
 for(i=0;i<closedlistep;++i)
  printf("%d[%d],",closedlist[i].nodeid,
                closedlist[i].parentid) ;
 printf("\n") ;
} 




