/************************************/
/*         横型探索プログラム       */
/*         width2.c                 */
/************************************/

#include <stdio.h>

/*   記号定数                      */
#define LIMITL 256 /*配列最大領域*/
#define TRUE 1 /*真（論理値）*/
#define START 1/*スタートの節点番号*/
#define GOAL 999 /*ゴールの節点番号*/

/*     節点を表現する構造体        */
struct node{
 int nodeid ;/*節点の番号*/
 int parentid ;/*接続元の節点番号*/
} ;
/* オープンリストとクローズドリスト*/
 struct node openlist[LIMITL] ;
 int openlistep=0 ;
 struct node closedlist[LIMITL] ;
 int closedlistep=0 ;

/*   関数のプロトタイプの宣言      */
/*リストの初期化*/
void initlist() ;
/*リストの状態の出力*/
void printlist() ;
/*節点の展開*/
void expand(int id) ;
/*先頭要素をクローズドリストへ移動*/
void movetofirst() ;
/*オープンリストの先頭から削除*/
void removefirst() ; 
/*ＩＤの重複チェック*/
int check(int id) ; 
/*ゴールへの経路を出力*/
void printroute(int id) ; 


/**********************/
/*     main()関数     */
/**********************/
int main()
{

 /*初期化*/
 initlist() ;
 printlist() ;
 /*探索の本体*/
 while(TRUE){
  if(openlistep==0){
   /*ゴールが見つからなかった*/
   printf("ゴールは見つかりませんでした\n") ;
   break ;
  }
  /*オープンリストの先頭がゴールなら終了*/
  if(openlist[0].nodeid==GOAL){
   printf("\nゴールを見つけました\n") ;
   printf("%d[%d]",openlist[0].nodeid,
                  openlist[0].parentid) ;
   printroute(openlist[0].parentid) ;
   break ;
  } 
  /*オープンリストの先頭要素を展開*/
  expand(openlist[0].nodeid) ;

  /*先頭要素をクローズドリストへ移動*/
   movetofirst() ;/*クローズドリストへ移動*/

  printlist() ;

 }
 
 return 0 ;
}
/**************************/
/*  printroute()関数      */
/*  ゴールへの経路を出力  */
/**************************/
void printroute(int id)  
{
 int i;
 /*節点番号idの接点を探す*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id){
   printf("<-%d[%d]",closedlist[i].nodeid,
                     closedlist[i].parentid) ;
   break ;
  }
 /*次の接点を出力*/
 if(closedlist[i].parentid!=0)
   /*スタートでなければ出力を続行*/ 
   printroute(closedlist[i].parentid) ;
 printf("\n") ;

}

/**************************/
/*     check()関数        */
/*  ＩＤの重複チェック    */
/**************************/
int check(int id)  
{
 int i;
 int res=0 ;

 /*オープンリストとの重複チェック*/
 for(i=0;i<openlistep;++i)
  if(openlist[i].nodeid==id)
   res=TRUE ;
 /*クローズドリストとの重複チェック*/
 for(i=0;i<closedlistep;++i)
  if(closedlist[i].nodeid==id)
   res=TRUE ;
 return res ;
} 


/**************************/
/*   removefirst()関数    */
/*オープンリストの先頭削除*/
/**************************/
void removefirst()
{
 int i;
 for(i=0;i<openlistep;++i)
  openlist[i]=openlist[i+1] ;
 --openlistep ;
} 


/**********************/
/*  movetofirst()関数 */
/*  先頭要素の移動    */
/**********************/
void movetofirst()
{
/*  先頭要素のクローズドリストへの移動    */
 closedlist[closedlistep++]=openlist[0]  ;
 /*オープンリストの先頭から削除*/
 removefirst() ; 
}



/**********************/
/*   expand()関数     */
/*   節点の展開       */
/**********************/
void expand(int id)
{
 /*接続関係の情報*/
 int tree[][5]={
  {1,2,3,4},
  {2,10,11},
  {3,12,13,14},
  {4,15,16,17},
  {11,21},
  {12,22},
  {14,23,24},
  {15,4,10,999},
  {16,25},
  {17,26,27},
  {0} /*情報の終わり*/
 } ;

 int i=0 ;
 int j ;

 while(tree[i][0]!=0){
  if(tree[i][0]==id){
   /*展開*/
   for(j=1;tree[i][j]!=0;++j){
    if(check(tree[i][j])!=TRUE){/*重複チェック*/
     openlist[openlistep].nodeid=tree[i][j] ;
     openlist[openlistep++].parentid=id ;
    }
   }
   break ;
  }
  ++i ;
 }

}


/**********************/
/*  initlist()関数    */
/*  リストの初期化    */
/**********************/
void initlist()
{
 /*オープンリストの初期化*/
 openlist[0].nodeid=START ;
 openlist[0].parentid=0 ;
 ++openlistep ;
}

/**********************/
/*  printlist()関数   */
/* リストの状態の出力 */
/**********************/
void printlist()
{
 int i ;

 /*オープンリストの出力*/
 printf("\nオープンリスト　　") ;
 for(i=0;i<openlistep;++i)
  printf("%d[%d],",openlist[i].nodeid,
                openlist[i].parentid) ;
 printf("\n") ;

 /*クローズドリストの出力*/
 printf("クローズドリスト　") ;
 for(i=0;i<closedlistep;++i)
  printf("%d[%d],",closedlist[i].nodeid,
                closedlist[i].parentid) ;
 printf("\n") ;



} 




