/************************************/
/*         ルールベール推論         */
/*         ifthen.c                 */
/*                                  */
/************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*   記号定数                      */
#define LIMITL 256 /*配列最大領域*/
#define TEXTLENGTH 4096 /* テキストの最大長*/
#define DATAFILENAME "ifdata.txt" /*データファイル*/
#define YES 1
#define NO  0

/*     ルールを表現する構造体        */
struct ifthen{
 char cond1[TEXTLENGTH] ;/*前件1*/
 char cond2[TEXTLENGTH] ;/*前件2*/
 char cond3[TEXTLENGTH] ;/*前件3*/
 char res[TEXTLENGTH]   ;/*後件*/
} ;

/*ルール集合*/
struct ifthen rulebase[LIMITL] ;
int rulepointer=0 ;/*ルール集合の大きさを格納*/

/*ワーキングメモリ*/
char wm[LIMITL][TEXTLENGTH] ;
int wmpointer=0 ;

/*   関数のプロトタイプの宣言      */
void initrule() ; /*ルールベースの初期化*/
void readdata() ; /*  データの読み込み  */
void setisalink(int ptr) ;/*リンクの生成*/
int searchb(char* B,int ptr) ; /*isalinkの探索*/
char* fgetl(char *text,int maxline,FILE *fp) ;
                  /*  ファイルの読み込み    */
void printrule() ;/*ルールベースの出力*/
int match(int i) ;/*ルールiとWMのマッチング*/
int singlematch(char *singlerule) ;
                  /*単独ルールとWMのマッチ  */
void putwm() ;/*    ＷＭの状態の出力    */

/**********************/
/*     main()関数     */
/**********************/
int main()
{
 char text[TEXTLENGTH] ;/*入力文字列*/
 int i ;
 int result ;/*検索結果*/
 int contflag ;/*続行フラグ*/
 int check[LIMITL]={NO} ;/*発火したルールの記録*/

 /*ルールの初期化*/
 initrule() ;
 printrule() ;

 /*探索の本体*/
 printf("「Ａ，Ｂ，・・・の特徴を持つのは何ですか」"
                         "という質問を扱います\n") ;
 printf("Ａ，Ｂ，・・・を改行で区切って入力してください．\n") ;
 while(fgets(text,TEXTLENGTH,stdin)!=NULL){
  if(strlen(text)<=1) break ;/*空行*/
  strncpy(wm[wmpointer++],text,strlen(text)-1) ;
 }
 printf("以下の特徴を持つものを調べます．\n") ;
 putwm() ;

 /*ルールベースの検索*/
 while(1){
  contflag=NO ;
  /*ルールを順に調べる*/
  for(i=0;i<rulepointer;++i){
   /*前件にマッチしたら後件をＷＭへ*/
   if((match(i)==YES)&&(check[i]==NO)){
    contflag=YES ;
    check[i]=YES ; 
    strcpy(wm[wmpointer++],rulebase[i].res) ;
    printf("WMの状態\n") ;
    putwm() ;
   }
  }

  
 /*ゴールが見つかったら終了*/
  if(strchr(wm[wmpointer-1],'<')!=NULL){
   printf(">%sを発見しました\n",wm[wmpointer-1]) ;
   break ;
  }
  else if(contflag==NO){
   printf("ゴールが見つかりませんでした\n") ;
   break ;
  }
 }

 return 0 ;
}

/**************************/
/*     putwm()関数        */
/*    ＷＭの状態の出力    */
/**************************/
void putwm()
{
 int i ;

 for(i=0;i<wmpointer;++i)
  printf("  %s\n",wm[i]) ;
}

/**************************/
/*  singlematch()関数     */
/*単独ルールとWMのマッチ  */
/**************************/
int singlematch(char *singlerule)
{
 int i ;
 int result=NO ;

 if(singlerule[0]=='*') return YES ;

 for(i=0;i<wmpointer;++i){
  if(strcmp(singlerule,wm[i])==0){ 
   result=YES ;
   break ;
  }
 }

 return result ;
}

/**************************/
/*  match()関数           */
/* ルールiとWMのマッチング*/
/**************************/
int match(int i)
{
 if((singlematch(rulebase[i].cond1)==YES) 
  &&(singlematch(rulebase[i].cond2)==YES) 
  &&(singlematch(rulebase[i].cond3)==YES))
  return YES ;/*マッチした*/

 else return NO ;/*マッチしない*/
}

/**************************/
/*  initrule()関数        */
/*  ルールベースの初期化  */
/**************************/
void initrule() 
{
 /*データの読み込み*/
 readdata() ;
}

/**************************/
/*     fgetl()関数        */
/*  ファイルの読み込み    */
/**************************/
char* fgetl(char *text,int maxline,FILE *fp) 
{
 while(fgets(text,maxline,fp) != NULL){
  if(strlen(text)>1)/*空行ではない*/
   return text ;
 }
 return NULL ; 
}

/**************************/
/*  readdata()関数        */
/*  データの読み込み      */
/**************************/
void readdata() 
{
 char c1text[TEXTLENGTH],
      c2text[TEXTLENGTH],
      c3text[TEXTLENGTH],
      rtext[TEXTLENGTH] ; /* 入力文字列*/
 FILE *fp ;/*ファイルポインタ*/

 /*ファイルオープン*/
 if((fp=fopen(DATAFILENAME,"r")) == NULL){
  /*ファイルオープン失敗*/
  printf("%sをオープンできません\n",DATAFILENAME) ;
  exit(1);
 }
 /*ファイルからの読み込み*/
 while((fgetl(c1text,TEXTLENGTH,fp) != NULL)
      &&(fgetl(c2text,TEXTLENGTH,fp) != NULL)
      &&(fgetl(c3text,TEXTLENGTH,fp) != NULL)
      &&(fgetl(rtext,TEXTLENGTH,fp) != NULL)){
   /*前件の代入*/
   strncpy(rulebase[rulepointer].cond1
                ,c1text,strlen(c1text)-1) ;
   strncpy(rulebase[rulepointer].cond2
                ,c2text,strlen(c2text)-1) ;
   strncpy(rulebase[rulepointer].cond3
                ,c3text,strlen(c3text)-1) ;
   /*後件の代入*/
   strncpy(rulebase[rulepointer].res
                ,rtext,strlen(rtext)-1) ;
   ++rulepointer ;
 }
 /*ファイルクローズ*/
 fclose(fp) ;
 
}

/**************************/
/*  printrule()関数       */
/*  debug function        */
/**************************/
void printrule() 
{
 int i ;
 for(i=0;i<rulepointer;++i){
  printf("rule %d\n",i) ;
  printf("cond1 : %s\n",rulebase[i].cond1) ;
  printf("cond2 : %s\n",rulebase[i].cond2) ;
  printf("cond3 : %s\n",rulebase[i].cond3) ;
  printf("res   : %s\n\n",rulebase[i].res) ;
 }
}
