/************************************/
/*         意味ネットワーク         */
/*         snet1.c                  */
/*                                  */
/*    その１ 「ＡはＢですか」       */
/************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*   記号定数                      */
#define LIMITL 256 /*配列最大領域*/
#define TEXTLENGTH 4096 /* テキストの最大長*/
#define DATAFILENAME "data.txt" /*データファイル*/
#define YES 1
#define NO  0

/*     節点を表現する構造体        */
struct node{
 char label[TEXTLENGTH] ;/*接点のラベル*/
 char isa[TEXTLENGTH] ;/*isaリンク*/
 int  isalink ; /*isaリンクの接続先番号*/
} ;

/*意味ネットワーク*/
struct node snet[LIMITL] ;
int snetpointer=0 ;/*snetの大きさを格納*/

/*   関数のプロトタイプの宣言      */
void initnet() ; /*ネットワークの初期化*/
void readdata() ; /*  データの読み込み  */
void setisalink(int ptr) ;/*リンクの生成*/
int searchb(char* B,int ptr) ; /*isalinkの探索*/
char* fgetl(char *text,int maxline,FILE *fp) ;
                  /*  ファイルの読み込み    */

/*デバッグ用関数*/
void printsnet()  ;/*意味ネットワークの表示*/

/**********************/
/*     main()関数     */
/**********************/
int main()
{
 char A[TEXTLENGTH],B[TEXTLENGTH] ;/*入力文字列*/
 int ptr ;
 int result ;/*検索結果*/

 /*ネットワークの初期化*/
 initnet() ;
 printsnet() ;/*意味ネットワークの表示*/

 /*探索の本体*/
 printf("「ＡはＢですか」という質問を扱います\n") ;
 printf("ＡとＢを改行で区切って入力してください．\n") ;
 while(fgets(A,TEXTLENGTH,stdin)!=NULL){
  if(fgets(B,TEXTLENGTH,stdin)!=NULL){
   if((strlen(A)<=1) || strlen(B)<=1)
        /*空行*/ break ;
   A[strlen(A)-1]='\0' ;
   B[strlen(B)-1]='\0' ;
  }
  printf("質問文「%sは%sですか」\n",A,B) ;
  /*Ａを探す*/
  for(ptr=0;ptr<=snetpointer;++ptr)
   if(strcmp(snet[ptr].label,A)==0)
    break ;
  if(ptr>snetpointer){
   printf("「%s」がわかりません\n",A) ;
   break ;
  }
  /*isaリンクをたどってＢを探す*/
  printf("  (考え中・・・)\n") ;
  result=searchb(B,ptr) ;
  printf("こたえ　「%sは%s",A,B) ;
  if(result==YES)
    printf("です．」\n") ;
  else
    printf("ではありません．」\n") ;
    
  printf("ＡとＢを改行で区切って入力してください．\n") ;
 }
 
 return 0 ;
}

/**************************/
/*  searchb()関数         */
/*  isalinkの探索         */
/**************************/
int searchb(char* B,int ptr) 
{
 printf("  (%sは%sです．)\n"
    ,snet[ptr].label,snet[snet[ptr].isalink].label) ;

 if(strcmp(B,snet[snet[ptr].isalink].label)==0)
  return YES ;
 else if(snet[snet[ptr].isalink].isalink!=-1)/*根ではない*/
  return searchb(B,snet[ptr].isalink) ;
 else return  NO ;
}

/**************************/
/*  setisalink()関数      */
/*  リンクの生成          */
/**************************/
void setisalink(int ptr) 
{
 int i ;

 for(i=0;i<=snetpointer;++i){
  if((i!=ptr)&&(strcmp(snet[ptr].isa
                        ,snet[i].label)==0)){
   snet[ptr].isalink=i ;
  }
 }
}

/**************************/
/*  initnet()関数         */
/*  ネットワークの初期化  */
/**************************/
void initnet() 
{
 int i ;

 /*データの読み込み*/
 snet[snetpointer].isalink=-1 ;/*根の設定*/
 readdata() ;
 /*リンクの生成*/
 for(i=0;i<=snetpointer;++i){
  setisalink(i) ;
 }
}

/**************************/
/*     fgetl()関数        */
/*  ファイルの読み込み    */
/**************************/
char* fgetl(char *text,int maxline,FILE *fp) 
{
 while(fgets(text,maxline,fp) != NULL){
  if(strlen(text)>1)/*空行ではない*/
   return text ;
 }
 return NULL ; 
}

/**************************/
/*  readdata()関数        */
/*  データの読み込み      */
/**************************/
void readdata() 
{
 char text[TEXTLENGTH],
      ltext[TEXTLENGTH] ; /* 入力文字列*/
 FILE *fp ;/*ファイルポインタ*/

 /*ファイルオープン*/
 if((fp=fopen(DATAFILENAME,"r")) == NULL){
  /*ファイルオープン失敗*/
  printf("%sをオープンできません\n",DATAFILENAME) ;
  exit(1);
 }
 /*ファイルからの読み込み*/
 while((fgetl(text,TEXTLENGTH,fp) != NULL)
      &&(fgetl(ltext,TEXTLENGTH,fp) != NULL)){
   /*リンク等の設定*/
  if(strncmp(text,"label",strlen("label"))==0){
    ++snetpointer ;
    strncpy(snet[snetpointer].label
                ,ltext,strlen(ltext)-1) ;
  }
  else if(strncmp(text,"isa",strlen("isa"))==0){
   strncpy(snet[snetpointer].isa
                ,ltext,strlen(ltext)-1) ;
  }
 }
 /*ファイルクローズ*/
 fclose(fp) ;
 
}

/**************************/
/*  printsnet()関数       */
/*  debug function        */
/**************************/
void printsnet() 
{
 int i ;
 for(i=0;i<=snetpointer;++i){
  printf("node %d\n",i) ;
  printf("label : %s\n",snet[i].label) ;
  printf("isa   : %s\n",snet[i].isa) ;
  printf("isalnk: %d\n",snet[i].isalink) ;
 }
}
