/************************************/
/*     tf-idfの計算プログラム       */
/*        tfidf.c                   */
/************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/*   記号定数                      */
#define NOFFILES 50 /*データファイル数上限*/
#define NOFWORDS 1000 /*単語数上限*/
#define WLENGTH 32 /*単語長の上限*/
#define LENGTH 256 /*文字列長の上限*/
#define EOD -2 /*データの終わりを示す記号*/

/*  単語の属性を表現する構造体        */
struct tuple{
 int freq ;/*出現頻度*/
 char word[WLENGTH] ;/*単語*/
 double tf ;
 double idf ;
} ;

/*関数のプロトタイプの宣言 */
void readfile(char *,struct tuple st[]) ;
               /* データ読み込み     */
int sumfreq(struct tuple st[]) ;
                /* 頻度合計の計算     */
void settf(struct tuple st[],int sumoffreq) ; 
                   /* settf()関数      */
void setidf(struct tuple t[][NOFWORDS]
                         ,int i,int argc)  ;
                    /* idf値の計算     */
int count(char word[]
    ,struct tuple t[][NOFWORDS],int argc)  ;
                     /* 単語出現ファイル数*/
void writefile(int no,struct tuple st[]) ;
                    /* データ書き込み     */


/**********************/
/*     main()関数     */
/**********************/
int main(int argc,char *argv[])
{
 int i;
 struct tuple t[NOFFILES][NOFWORDS] ;
 int sumoffreq[NOFFILES] ;/*頻度の合計*/

 /*コマンドライン引数のチェック*/
 if(argc<3){/*引数が足りない*/
  fprintf(stderr,"使用法　tfidf "
         "(ファイル1) (ファイル2) ・・・\n") ;
  exit(1) ;
 }
 else if(argc>NOFFILES){/*ファイルが多すぎる*/
  fprintf(stderr,"ファイルが多すぎます（%d)\n "
         ,argc) ;
  exit(1) ;
 }


 /*ファイルからのデータ読み込み*/
 for(i=0;i<(argc-1);++i){
  readfile(argv[i+1],t[i]) ;
 }

 /*頻度合計の計算*/
 for(i=0;i<(argc-1);++i){
  sumoffreq[i]=sumfreq(t[i]) ; 
 }

 /*tf値の計算*/
 for(i=0;i<(argc-1);++i){
  settf(t[i],sumoffreq[i]) ; 
 }

 /*idf値の計算*/
 for(i=0;i<(argc-1);++i){
  setidf(t,i,argc) ; 
 }
 
 /*ファイルへのデータ書き込み*/
 for(i=0;i<(argc-1);++i){
  writefile(i,t[i]) ;
 }

 return 0 ;
}

/**********************/
/* writefile()関数    */
/* データ書き込み     */
/**********************/
void writefile(int no,struct tuple st[])
{
 FILE *fp ;/*ファイルポインタ*/
 char filename[LENGTH] ;/*ファイル名*/
 int i=0 ;

 /*ファイル名は数字とする*/
 sprintf(filename,"%d.txt",no) ;

 /*ファイルオープン*/
 if((fp=fopen(filename,"w"))==NULL){
  /*ファイルオープンのエラー*/
  printf("%sが開けません\n",filename) ;
  exit(1) ;
 } 

 /*ファイルへの書き込み*/
  while(st[i].freq!=EOD){
   fprintf(fp,"%lf %s\n",st[i].tf*st[i].idf
                                 ,st[i].word) ;
    ++ i ;
  }

 /*ファイルクローズ*/
 fclose(fp) ;
}

/**********************/
/* setidf()関数       */
/* idf値の計算        */
/**********************/
void setidf(struct tuple t[][NOFWORDS]
                       ,int no,int argc) 
{
 int i=0 ;

 while(t[no][i].freq!=EOD){
  t[no][i].idf=log(argc/
    (double)count(t[no][i].word,t,argc)) ;
  ++i ;
 }
}

/**********************/
/*  count()関数       */
/* 単語出現ファイル数 */
/**********************/
int count(char word[]
    ,struct tuple t[][NOFWORDS],int argc) 
{
 int cnt=0 ;/*出現数*/
 int i,j ;

 for(i=0;i<argc-1;++i){
  j=0 ;
  while(t[i][j].freq!=EOD){
   if(strcmp(word,t[i][j].word)==0){
    ++cnt ;
    break ;
   }
   ++j ;
  }
 }
 return cnt ;
}


/**********************/
/* settf()関数        */
/* tf値の計算         */
/**********************/
void settf(struct tuple st[],int sumoffreq) 
{
 int i=0 ;

 while(st[i].freq!=EOD){
  st[i].tf=st[i].freq/(double)sumoffreq ;
  ++ i ;
 }
}

/**********************/
/* sumfreq()関数      */
/* 頻度合計の計算     */
/**********************/
int sumfreq(struct tuple st[])
{
 int i=0 ;
 int sum=0 ;

 while(st[i].freq!=EOD){
  sum+=st[i].freq ;
  ++ i ;
 }
 
 return sum ;
}

/**********************/
/* readfile()関数     */
/* データ読み込み     */
/**********************/
void readfile(char *filename,struct tuple st[])
{
 int i=0 ;
 FILE *fp ;/*ファイルポインタ*/
 char buffer[LENGTH] ;/*入力バッファ*/

 /*ファイルオープン*/
 if((fp=fopen(filename,"r"))==NULL){
  /*ファイルオープンのエラー*/
  printf("%sが開けません\n",filename) ;
  exit(1) ;
 } 

 /*ファイルからの読み込み*/
 while(fgets(buffer,LENGTH,fp)!=NULL){
  if(sscanf(buffer ,"%d %s",&(st[i].freq)
              ,st[i].word)!=0)
   /*変換できた*/
  if(strlen(st[i].word)!=0)
   ++ i;/*次の位置へ*/
  if(i>=(NOFWORDS-1)){/*単語数の上限を超えた*/
   fprintf(stderr,"単語数が多すぎます "
         "(%s)\n",filename) ;
   break ;/*読み込みの打ち切り*/ 
  }
 }
 /*ファイルの終わりの記号*/
 st[i].freq=EOD ;

 /*ファイルクローズ*/
 fclose(fp) ;
}

